/*  Copyright 2004-2009, Unpublished Work of Technologic Systems
 *  All Rights Reserved.
 *
 *  THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
 *  PROPRIETARY AND TRADE SECRET INFORMATION OF TECHNOLOGIC SYSTEMS.
 *  ACCESS TO THIS WORK IS RESTRICTED TO (I) TECHNOLOGIC SYSTEMS EMPLOYEES
 *  WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE OF THEIR
 *  ASSIGNMENTS  AND (II) ENTITIES OTHER THAN TECHNOLOGIC SYSTEMS WHO
 *  HAVE ENTERED INTO  APPROPRIATE LICENSE AGREEMENTS.  NO PART OF THIS
 *  WORK MAY BE USED, PRACTICED, PERFORMED, COPIED, DISTRIBUTED, REVISED,
 *  MODIFIED, TRANSLATED, ABRIDGED, CONDENSED, EXPANDED, COLLECTED,
 *  COMPILED,LINKED,RECAST, TRANSFORMED, ADAPTED IN ANY FORM OR BY ANY
 *  MEANS,MANUAL, MECHANICAL, CHEMICAL, ELECTRICAL, ELECTRONIC, OPTICAL,
 *  BIOLOGICAL, OR OTHERWISE WITHOUT THE PRIOR WRITTEN PERMISSION AND
 *  CONSENT OF TECHNOLOGIC SYSTEMS . ANY USE OR EXPLOITATION OF THIS WORK
 *  WITHOUT THE PRIOR WRITTEN CONSENT OF TECHNOLOGIC SYSTEMS  COULD
 *  SUBJECT THE PERPETRATOR TO CRIMINAL AND CIVIL LIABILITY.
 */
#include <assert.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <linux/unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <time.h>
#include <limits.h>
#include <sched.h>
#include <libgen.h>
#include <time.h>

#define DIOMASK 0x0fffffff07ff7fffULL

#define PMUA_BASE_PHYSICAL 0xD4280000     /* a.k.a. APMU, Application Subsystem Power Management (-0x2800!!!) */
#define MPMU_BASE_PHYSICAL 0xD4050000     /* Main Power Management Unit, a.k.a. PMUM */

static void calculateClocks(unsigned int *pclk, unsigned int *dclk);
static volatile unsigned short *syscon;
static void *mpmu_base;
static void *pmua_base;

static inline unsigned short PEEK16(unsigned long addr) {
        unsigned short ret;

        asm volatile (
                "ldrh %0, [ %1 ]\n"
                : "=r" (ret)
                : "r" (addr)
                : "memory"
        );
        return ret;
}

static inline void POKE16(unsigned long addr, unsigned short dat) {
        asm volatile (
                "strh %1, [ %0 ]\n"
                :
                : "r" (addr), "r" (dat)
                : "memory"
        );
}

#define peek16(adr) PEEK16((unsigned long)&syscon[(adr)/2])
#define poke16(adr, val) POKE16((unsigned long)&syscon[(adr)/2],(val))


void sspi_cmd(unsigned int cmd) {
  unsigned int i;
  unsigned int s = peek16(0x28) & 0xfff0;

  // pulse CS#
  poke16(0x28, s | 0x2);
  poke16(0x28, s | 0x0);

  for (i = 0; i < 32; i++, cmd <<= 1) {
    if (cmd & 0x80) {
      poke16(0x28, s | 0x4);
      poke16(0x28, s | 0xc);
    } else {
      poke16(0x28, s | 0x0);
      poke16(0x28, s | 0x8);
    }
  }
}

static int read_tagmem(unsigned int *tagmem) {
  int i, j;
  unsigned int ret;
  unsigned int s = peek16(0x28) & 0xfff0;

  sspi_cmd(0xac); // X_PROGRAM_EN
  sspi_cmd(0x4e); // READ_TAG

  for (j = 0; j < 20; j++) {
    for (ret = 0x0, i = 0; i < 32; i++) {
      poke16(0x28, s | 0x0);
      poke16(0x28, s | 0x8);
      ret = ret << 1 | (peek16(0x28) & 0x1);
    }
    tagmem[j] = ret;
  }

  sspi_cmd(0x78); // PROGRAM_DIS

  poke16(0x28, s | 0x2);
  return 1;
}

static int write_tagmem(unsigned int *tagmem) {
  int i, j;
  unsigned int s = peek16(0x28) & 0xfff0;

  sspi_cmd(0xac); // X_PROGRAM_EN
  sspi_cmd(0x8e); // WRITE_TAG

  for (j = 0; j < 20; j++) {
    unsigned int x = tagmem[j];
    for (i = 0; i < 32; i++, x <<= 1) {
      if (x & 0x80000000UL) {
	poke16(0x28, s | 0x4);
	poke16(0x28, s | 0xc);
      } else {
	poke16(0x28, s | 0x0);
	poke16(0x28, s | 0x8);
      }
      if (i == 23 && j == 19) break;
    }
  }

  for (i = 0; i < 8; i++) {
    poke16(0x28, s | 0x2);
    poke16(0x28, s | 0xa);
  }
  poke16(0x28, s | 0x2);
  usleep(25000);
  return 1;
}

int erase_tagmem() {
  int i;
  unsigned int s = peek16(0x28) & 0xfff0;

  sspi_cmd(0xac); // X_PROGRAM_EN
  sspi_cmd(0xe); // ERASE_TAG

  for (i = 0; i < 8; i++) {
    poke16(0x28, s | 0x2);
    poke16(0x28, s | 0xa);
  }
  poke16(0x28, s | 0x2);
  usleep(1000000);
  return 1;
}

static void usage(char **argv) {
	fprintf(stderr, "Usage: %s [OPTION] ...\n"
	  "Technologic Systems TS-4700 FPGA manipulation.\n"
	  "\n"
	  "General options:\n"
	  "  -g, --getmac            Display ethernet MAC address\n"
	  "  -s, --setmac=MAC        Set ethernet MAC address\n"
	  "  -R, --reboot            Reboot the board\n"
	  //"  -t, --getrtc            Display RTC time/date\n"
	  //"  -S, --setrtc            Set RTC time/date from system time\n"
	  "  -i, --info              Display board FPGA info\n"
	  "  -B, --baseboard         Display baseboard ID\n"
	  "  -a, --adc               Display MCP3428 ADC readings in millivolts\n"
	  "  -e, --greenledon        Turn green LED on\n"
	  "  -b, --greenledoff       Turn green LED off\n"
	  "  -c, --redledon          Turn red LED on\n"
	  "  -d, --redledoff         Turn red LED off\n"
	  "  -D, --setdio=LVAL       Set DIO output to LVAL\n"
	  "  -O, --setdiodir=LVAL    Set DIO direction to LVAL (1 - output)\n"
	  "  -G, --getdio            Get DIO input\n"
	  "  -Z, --getdioreg         Get DIO direction and output register values\n"
	  "  -x, --random            Get 16-bit hardware random number\n"
	  "  -W, --watchdog          Daemonize and set up /dev/watchdog\n"
	  "  -A, --autofeed=SETTING  Daemonize and auto feed watchdog\n"
	  "  -n, --setrng            Seed the kernel random number generator\n"
	  "  -X, --resetswitchon     Enable reset switch\n"
	  "  -Y, --resetswitchoff    Disable reset switch\n"
	  "  -l, --loadfpga=FILE     Load FPGA bitstream from FILE\n"
	  "  -k  --clocks            Display the CPU/DRAM clock speeds from FPGA tagmem\n"
	  //"  -u  --getborn           Display the 'born-on' timestamp\n"
	  //"  -U  --setborn           Set the 'born-on' timestamp (same format as the 'date' command)\n"
	  "  -h, --help              This help\n",
	  argv[0]
	);
}

static inline
unsigned char tobcd(unsigned int n) {
	unsigned char ret;
	ret = (n / 10) << 4;
	ret |= n % 10;
	return ret;
}

static inline unsigned int frombcd(unsigned char x) {
	unsigned int ret;
	ret = (x >> 4) * 10;
	ret += x & 0xf;
	return ret;
}

int main(int argc, char **argv) {
	int devmem, i, c, wdog;
	unsigned int tagmem[20];
	unsigned int baseboard = 0;
	int opt_getmac = 0, opt_setmac = 0, opt_reboot = 0, opt_getrtc = 0;
	int opt_setrtc = 0, opt_info = 0, opt_greenledon = 0;
	int opt_greenledoff = 0, opt_redledon = 0, opt_redledoff = 0;
	int opt_setdio = 0, opt_setdiodir = 0, opt_getdio = 0, opt_random = 0;
	int opt_watchdog = 0, opt_autofeed = 0, opt_resetswitchon = 0;
	int opt_resetswitchoff = 0;
	int opt_getdioreg = 0, opt_setrng = 0;
	int opt_baseboard = 0, opt_adc = 0, opt_clocks = 0, opt_setclocks=0;
	int opt_getborn=0, opt_setborn=0, bornon=0;
	unsigned long long dio, oe;
	char *opt_mac, *opt_loadfpga = NULL;
	static struct option long_options[] = {
	  { "help", 0, 0, 'h' },
	  { "getmac", 0, 0, 'g' },
	  { "setmac", 1, 0, 's' },
	  { "reboot", 0, 0, 'R' },
	  { "getrtc", 0, 0, 't' },
	  { "setrtc", 0, 0, 'S' },
	  { "greenledon", 0, 0, 'e'},
	  { "greenledoff", 0, 0, 'b'},
	  { "redledon", 0, 0, 'c'},
	  { "redledoff", 0, 0, 'd'},
	  { "setdio", 1, 0, 'D'},
	  { "setdiodir", 1, 0, 'O'},
	  { "getdio", 0, 0, 'G'},
	  { "getdioreg", 0, 0, 'Z'},
	  { "info", 0, 0, 'i' },
	  { "baseboard", 0, 0, 'B' },
	  { "adc", 0, 0, 'a' },
	  { "random", 0, 0, 'x' },
	  { "watchdog", 0, 0, 'W' },
	  { "autofeed", 1, 0, 'A' },
	  { "setrng", 0, 0, 'n' },
	  { "resetswitchon", 0, 0, 'X' },
	  { "resetswitchoff", 0, 0, 'Y' },
	  { "loadfpga", 1, 0, 'l' },
	  { "clocks", 2, 0, 'k' },
	  { "getborn", 0, 0, 'u' },
     { "setborn", 1, 0, 'U' },
	  { 0, 0, 0, 0 }
	};


	while((c = getopt_long(argc, argv,
	  "hgs:RtSiBaebcdD:O:GxWA:XYZl:nk::uU:", long_options, NULL)) != -1) {
		switch (c) {
		case 'n':
			opt_setrng = 1;
			break;
		case 'l':
			opt_loadfpga = strdup(optarg);
			break;
		case 'k':
		   opt_clocks = 1;
		   if (optarg) {
		      opt_setclocks = atoi(optarg);
		      if (opt_setclocks < 1 || opt_setclocks > 5) {
		         fprintf(stderr, "clock must be between 1 and 5\n");
		         exit(1);
		      }
		   }
		   break;
		case 'Z':
			opt_getdioreg = 1;
			break;
		case 'X':
			opt_resetswitchon = 1;
			break;
		case 'Y':
			opt_resetswitchoff = 1;
			break;
		case 'W':
			opt_watchdog = 1;
			break;
		case 'A':
			opt_watchdog = 1;
			opt_autofeed = 1;
			wdog = strtoul(optarg, NULL, 0);
			if (wdog > 3) {
				fprintf(stderr,
				  "autofeed must be between 0 and 3\n");
				exit(1);
			}
			break;
		case 'D':
			opt_setdio = 1;
			dio = strtoull(optarg, NULL, 0);
			break;
		case 'O':
			opt_setdiodir = 1;
			oe = strtoull(optarg, NULL, 0);
			break;
		case 'G':
			opt_getdio = 1;
			break;
		case 'e':
			opt_greenledon = 1;
			break;
		case 'b':
			opt_greenledoff = 1;
			break;
		case 'c':
			opt_redledon = 1;
			break;
		case 'd':
			opt_redledoff = 1;
			break;
		case 'i':
			opt_info = 1;
			opt_baseboard = 1;
			break;
		case 'B':
			opt_baseboard = 1;
			break;
		case 'a':
			opt_adc = 1;
			opt_baseboard = 1;
			break;
		case 'S':
			opt_setrtc = 1;
			break;
		case 'g':
			opt_getmac = 1;
			break;
		case 's':
			opt_setmac = 1;
			opt_mac = strdup(optarg);
			break;
		case 'R':
			opt_reboot = 1;
			break;
		case 't':
			opt_getrtc = 1;
			break;
		case 'x':
			opt_random = 1;
			break;
		case 'u':
		   opt_getborn = 1;
		   break;
		case 'U':
		   opt_setborn = 1;
		   bornon = strtoull(optarg, NULL, 0);
		   break;
		case 'h':
		default:
			usage(argv);
			return(1);
		}
	}


	devmem = open("/dev/mem", O_RDWR|O_SYNC);
	assert(devmem != -1);
	syscon = (unsigned short *) mmap(0, 4096,
	  PROT_READ | PROT_WRITE, MAP_SHARED, devmem, 0x80004000);

	mpmu_base = mmap(0, 4096,
     PROT_READ | PROT_WRITE, MAP_SHARED, devmem, MPMU_BASE_PHYSICAL);

	pmua_base = mmap(0, 4096+0x2800,
     PROT_READ | PROT_WRITE, MAP_SHARED, devmem, PMUA_BASE_PHYSICAL);

	//if (opt_loadfpga) {
		//unsigned int prev_gpioa_en, prev_gpioa_ddr;
		//signed char x;
		//const char * ispvmerr[] = { "pass", "verification fail",
		  //"can't find the file", "wrong file type", "file error",
		  //"option error", "crc verification error" };


        	//cvmiscregs = (unsigned int *) mmap(0, 4096,
		  //PROT_READ | PROT_WRITE, MAP_SHARED, devmem, 0x76000000);
		//cvtimerregs = (unsigned int *) mmap(0, 4096,
		  //PROT_READ | PROT_WRITE, MAP_SHARED, devmem, 0x79000000);
		//prev_gpioa_en = cvmiscregs[0x20/4];
		//prev_gpioa_ddr = cvgpioregs[0x8/4];
		//cvmiscregs[0x20/4] &= ~0x1c00004; /* switch A2, A22-A24 into GPIO mode */
		//cvgpioregs[0x8/4] |= (1<<22|1<<23|1<<2); /* A2, A22, A23 as outputs */
		//cvgpioregs[0x8/4] &= ~(1<<24); /* A24 as inputs */

		//x = ispVM(opt_loadfpga);
		//if (x == 0) {
			//printf("loadfpga_ok=1\n");
		//} else {
			//assert(x < 0);
			//printf("loadfpga_ok=0\n");
			//printf("loadfpga_error=\"%s\"\n", ispvmerr[-x]);
		//}

		//cvmiscregs[0x20/4] = prev_gpioa_en;
		//cvgpioregs[0x8/4] = prev_gpioa_ddr;
	//} else sbuslock();

	if (opt_random) printf("random=0x%x%x\n", peek16(0x0e), peek16(0x0c));
	if (opt_setrng) {
		FILE *urandom = NULL;
		unsigned int rng;
		rng = peek16(0x0e) << 16;
		rng |= peek16(0xc);

		urandom = fopen("/dev/urandom", "w");
		assert(urandom != NULL);
		fwrite(&rng, 2, 1, urandom);
		fclose(urandom);
	}

	if (opt_resetswitchon) {
		unsigned short x;
		x = peek16(0x2);
		x |= 0x8000;
		poke16(0x2, x);
	} else if (opt_resetswitchoff) {
		unsigned short x;
		x = peek16(0x2);
		x &= ~0x8000;
		poke16(0x2, x);
	}
/*

	if (opt_extendedtempon) {
		read_tagmem(tagmem);
		erase_tagmem();
		tagmem[0] &= ~0x80;
		write_tagmem(tagmem);
	} else if (opt_extendedtempoff) {
		read_tagmem(tagmem);
		erase_tagmem();
		tagmem[0] |= 0x80;
		write_tagmem(tagmem);
	}
*/
	if (opt_info) {
		unsigned short x;
		unsigned int pclk, dclk;
		printf("model=0x%x\n", peek16(0x00));
		x = peek16(0x02);
		printf("submodel=0x%x\n", peek16(0x2a));
		printf("revision=0x%x\n", x & 0xf);
		printf("bootmode=0x%x\n", (x >> 4) & 0x3);
		printf("bootdev=0x%x\n", (x >> 6) & 0x3);
		//printf("resetsw_en=0x%x\n", (x >> 4) & 0x1);
		//printf("latched_dio9=0x%x\n", (x >> 5) & 0x1);
		opt_getdio = 1;

		read_tagmem(tagmem);
      printf("bornon=%d\n", tagmem[1]);

      calculateClocks(&pclk, &dclk);

      printf("pclk=%d\ndclk=%d\n", pclk, dclk);

	}

	if (opt_getdio) {
		unsigned long long v;
		v = peek16(0x20);
		v |= (unsigned long long)peek16(0x22) << 16;
		v |= (unsigned long long)peek16(0x24) << 32;
		v |= (unsigned long long)peek16(0x26) << 48;
		v &= DIOMASK;
		printf("dio=0x%llx\n", v);
	}

	if (opt_setdio) {
		unsigned int leds;

		leds = peek16(0x12);
		leds = leds << 16;
		leds &= 0x18000000;
		dio &= DIOMASK;
		dio |= leds;
		poke16(0x10, dio & 0xffff);
		poke16(0x12, (dio >> 16) & 0xffff);
		poke16(0x14, (dio >> 32) & 0xffff);
		poke16(0x16, (dio >> 48) & 0xffff);
	}

	if (opt_setdiodir) {
		oe &= DIOMASK;
		poke16(0x18, oe & 0xffff);
		poke16(0x1a, (oe >> 16) & 0xffff);
		poke16(0x1c, (oe >> 32) & 0xffff);
		poke16(0x1e, (oe >> 48) & 0xffff);
	}

	if (opt_getdioreg) {
		unsigned long long v;
		v = peek16(0x18);
		v |= (unsigned long long)peek16(0x1a) << 16;
		v |= (unsigned long long)peek16(0x1c) << 32;
		v |= (unsigned long long)peek16(0x1e) << 48;
		v &= DIOMASK;
		printf("diodir=0x%llx\n", v);
		v = peek16(0x10);
		v |= (unsigned long long)peek16(0x12) << 16;
		v |= (unsigned long long)peek16(0x14) << 32;
		v |= (unsigned long long)peek16(0x16) << 48;
		v &= DIOMASK;
		printf("dio_out=0x%llx\n", v);
	}

	if (opt_baseboard) {
		unsigned short prev1, prev2, prev3, prev4;
		unsigned int i, x;

		prev1 = peek16(0x4);
		prev2 = peek16(0x12);
		prev3 = peek16(0x1a);
		prev4 = peek16(0x10);
		poke16(0x4, 0); // disable muxbus
		poke16(0x10, prev4 & ~0x20);
		poke16(0x1a, prev3 | 0x200);
		for(i=0; i<8; i++) {
			x = prev2 & ~0x1a00;
			if (!(i & 1)) x |= 0x1000;
			if (!(i & 2)) x |= 0x0800;
			if (i & 4) x |= 0x0200;
			poke16(0x12, x);
			usleep(1);
			baseboard = (baseboard >> 1);
			if (peek16(0x20) & 0x20) baseboard |= 0x80;
		}
		poke16(0x4, prev1);
		poke16(0x12, prev2);
		poke16(0x1a, prev3);
		poke16(0x10, prev4);
		printf("baseboard_model=%x\n", baseboard);
		switch (baseboard) {
			case 2:
				printf("baseboard=8390\n");
				break;
			case 6:
				printf("baseboard=8160\n");
				break;
			case 7:
				printf("baseboard=8100\n");
				break;
			case 255:
				printf("baseboard=8200\n");
				break;
		}

	}

	if (opt_adc) {
		int x;
		// configure ADC:
		switch (baseboard) {
			case 2:
				poke16(0x400, 0x28);
				break;
			case 6:
			case 7:
				poke16(0x400, 0x18);
				break;
			default: // if unknown baseboard, uses no an_sel
				// but assumes ADC is there
				poke16(0x400, 0x08);
				break;
		}
		poke16(0x2, peek16(0x2) | 0x800); // enable ADC core
		poke16(0x402, 0x3f); // enable all 6 channels
		usleep(500000); // allow time for conversions
		for (i = 1; i <= 6; i++) {
			x = (signed short)peek16(0x402 + 2*i);
			if (i > 2) x = (x * 1006)/200;
			x = (x * 2048)/0x8000;
			printf("adc%d=%d\n", i, x);
		}
	}

	if (opt_clocks) {
      unsigned int c;
      read_tagmem(tagmem);
      if (opt_setclocks) {
         c = tagmem[0] & ~7;
         printf("Speed set to ");
         switch(opt_setclocks) {
            case 1: printf("Core: 156MHz  SDRAM: 156MHz\n"); break;           // 000b
            case 2: c |= 7; printf("Core: 624MHz  SDRAM: 312MHz\n"); break;   // 111b
            case 3: c |= 1; printf("Core: 800MHz  SDRAM: 400MHz\n"); break;   // 001b
            case 4: c |= 2; printf("Core: 1066MHz  SDRAM: 355MHz\n"); break;  // 010b
            case 5: c |= 3; printf("Core: 1200MHz  SDRAM: 400MHz\n"); break;  // 011b
         }
         tagmem[0] = c;
         erase_tagmem();
         write_tagmem(tagmem);

      } else {
         switch(tagmem[0] & 7) {
            case 0: printf("Core: 156MHz  SDRAM: 156MHz\n"); break;
            case 7: printf("Core: 624MHz  SDRAM: 312MHz\n"); break;
            case 1: printf("Core: 800MHz  SDRAM: 400MHz\n"); break;
            case 2: printf("Core: 1066MHz  SDRAM: 355MHz\n"); break;
            case 3: printf("Core: 1200MHz  SDRAM: 400MHz\n"); break;
         }
      }
   }

	if (opt_getborn) {
	    read_tagmem(tagmem);
	    printf("%d\n", tagmem[1]);
	}

	if (opt_setborn) {
	    read_tagmem(tagmem);
	    tagmem[1] = bornon;
	    erase_tagmem();
       write_tagmem(tagmem);
	}

	if (opt_getmac) {
		unsigned char a, b, c;
		read_tagmem(tagmem);
		a = tagmem[0] >> 24;
		b = tagmem[0] >> 16;
		c = tagmem[0] >> 8;
		printf("mac=00:d0:69:%02x:%02x:%02x\n", a, b, c);
	}

	if (opt_setmac) {
		unsigned int a, b, c;
		int r;
		r = sscanf(opt_mac, "%*x:%*x:%*x:%x:%x:%x",  &a, &b, &c);
		assert(r == 3); /* XXX: user arg problem */
		assert(a < 0x100);
		assert(b < 0x100);
		assert(c < 0x100);
		read_tagmem(tagmem);
		erase_tagmem();
		tagmem[0] &= ~0xffffff00;
		tagmem[0] |= (a<<24|b<<16|c<<8);
		write_tagmem(tagmem);
	}

	if (opt_reboot) {
		struct sched_param sched;
		sync();
		poke16(0x06, 0);
		sched.sched_priority = 99;
		sched_setscheduler(0, SCHED_FIFO, &sched);
		while(1);
	}

	if (opt_greenledoff) poke16(0x12, peek16(0x12) & ~0x800);
	if (opt_greenledon) poke16(0x12, peek16(0x12) | 0x800);
	if (opt_redledoff) poke16(0x12, peek16(0x12) & ~0x1000);
	if (opt_redledon) poke16(0x12, peek16(0x12) | 0x1000);

//	if (opt_setrtc) {
//		unsigned short x, led;
//		unsigned int d, ack;
//		unsigned char rtcdat[9];
//		time_t now = time(NULL);
//		struct tm *tm;
//
//		now += 6; /* 6 seconds in future to accomodate sleep() below */
//		tm = gmtime(&now);
//		rtcdat[0] = 0;
//		rtcdat[1] = tobcd(tm->tm_sec) | 0x80;
//		rtcdat[2] = tobcd(tm->tm_min) | 0x80;
//		rtcdat[3] = tobcd(tm->tm_hour);
//		rtcdat[4] = tm->tm_wday + 1;
//		rtcdat[5] = tobcd(tm->tm_mday);
//		rtcdat[6] = tobcd(tm->tm_mon + 1);
//		rtcdat[7] = tobcd(tm->tm_year % 100);
//		rtcdat[8] = 0;
//
//		led = peek16(0x62) & 0xc000;
//resettime:
//		poke16(0x62, led|0x0); /* tristate, scl/sda pulled up */
//		poke16(0x62, led|0xe00); /* i2c, start (sda low) */
//		x = 0xe00|led;
//		for (d = 0xd0, i = 0; i < 8; i++, d <<= 1) {
//			x &= ~0x200;
//			poke16(0x62, x); /* scl low */
//			if (d & 0x80) x |= 0x100;
//			else x &= ~0x100;
//			poke16(0x62, x); /* sda update */
//			x |= 0x200;
//			poke16(0x62, x); /* scl high */
//		}
//		poke16(0x62, led|0x800); /* scl low, tristate sda */
//		poke16(0x62, led|0xa00); /* scl high, tristate sda */
//		ack = (peek16(0x62) >> 12) & 0x1; /* sample ack */
//		if (ack) {
//			fprintf(stderr, "rtc_present=0\n");
//			goto skiprtc;
//		}
//
//		for (j = 0; j < 9; j++) {
//			for (d = rtcdat[j], i = 0; i < 8; i++, d <<= 1) {
//				x &= ~0x200;
//				poke16(0x62, x); /* scl low */
//				if (d & 0x80) x |= 0x100;
//				else x &= ~0x100;
//				poke16(0x62, x); /* sda update */
//				x |= 0x200;
//				poke16(0x62, x); /* scl high */
//			}
//			poke16(0x62, led|0x800); /* scl low, tristate sda */
//			poke16(0x62, led|0xa00); /* scl high, tristate sda */
//			ack = (peek16(0x62) >> 12) & 0x1; /* sample ack */
//			assert(ack == 0);
//		}
//
//		poke16(0x62, led|0xc00); /* scl low, sda low */
//		poke16(0x62, led|0xe00); /* scl high, sda low */
//		poke16(0x62, led|0x0); /* tristate (i2c stop) */
//		if (rtcdat[1] & 0x80) {
//			rtcdat[1] &= ~0x80;
//			goto resettime;
//		}
//		if (rtcdat[2] & 0x80) {
//			sleep(6);
//			rtcdat[2] &= ~0x80;
//			goto resettime;
//		}
//		fprintf(stderr, "rtc_present=1\n");
//
//	}
//
//	if (opt_getrtc) {
//		unsigned short x, led;
//		unsigned int d, ack;
//		unsigned char rtcdat[8];
//		time_t now;
//		struct tm tm;
//		struct timeval tv;
//
//		led = peek16(0x62) & 0xc000;
//		poke16(0x62, led|0x0); /* tristate, scl/sda pulled up */
//		poke16(0x62, led|0xe00); /* i2c, start (sda low) */
//		poke16(0x62, led|0xf00); /* i2c, stop (sda high) */
//		poke16(0x62, led|0xe00); /* i2c, start (sda low) */
//		x = led|0xe00;
//		for (d = 0xd0, i = 0; i < 8; i++, d <<= 1) {
//			x &= ~0x200;
//			poke16(0x62, x); /* scl low */
//			if (d & 0x80) x |= 0x100;
//			else x &= ~0x100;
//			poke16(0x62, x); /* sda update */
//			x |= 0x200;
//			poke16(0x62, x); /* scl high */
//		}
//		poke16(0x62, led|0x800); /* scl low, tristate sda */
//		poke16(0x62, led|0xa00); /* scl high, tristate sda */
//		ack = (peek16(0x62) >> 12) & 0x1; /* sample ack */
//		fprintf(stderr, "rtc_present=%d\n", ack ? 0 : 1);
//		if (ack) goto skiprtc;
//		for (i = 0; i < 8; i++) {
//			poke16(0x62, led|0xc00); /* scl low, sda low */
//			poke16(0x62, led|0xe00); /* scl high */
//		}
//		poke16(0x62, led|0x800); /* scl low, tristate sda */
//		poke16(0x62, led|0xa00); /* scl high, tristate sda */
//		ack = (peek16(0x62) >> 12) & 0x1; /* sample ack */
//		assert(ack == 0);
//		poke16(0x62, led|0x800); /* scl low, tristate sda */
//		poke16(0x62, led|0xd00); /* scl low, sda high */
//		poke16(0x62, led|0xf00); /* scl high, sda high */
//		poke16(0x62, led|0xe00); /* scl high, sda low (repeated start) */
//		x = led|0xe00;
//		for (d = 0xd1, i = 0; i < 8; i++, d <<= 1) {
//			x &= ~0x200;
//			poke16(0x62, x); /* scl low */
//			if (d & 0x80) x |= 0x100;
//			else x &= ~0x100;
//			poke16(0x62, x); /* sda update */
//			x |= 0x200;
//			poke16(0x62, x); /* scl high */
//		}
//		poke16(0x62, led|0x800); /* scl low, tristate sda */
//		poke16(0x62, led|0xa00); /* scl high, tristate sda */
//		ack = (peek16(0x62) >> 12) & 0x1; /* sample ack */
//		assert(ack == 0);
//		for (j = 0; j < 8; j++) {
//			for (i = 0; i < 8; i++) {
//				poke16(0x62, led|0x800); /* scl low, tristate sda */
//				poke16(0x62, led|0xa00); /* scl high, tristate sda */
//				rtcdat[j] <<= 1;
//				rtcdat[j] |= (peek16(0x62) >> 12) & 0x1;
//			}
//			if (j != 7) {
//				poke16(0x62, led|0x800); /* scl low, tristate sda */
//				poke16(0x62, led|0xc00); /* scl low, sda low */
//				poke16(0x62, led|0xe00); /* scl high, sda low */
//			} else {
//				poke16(0x62, led|0x800); /* scl low, tristate sda */
//				poke16(0x62, led|0xd00); /* scl low, sda high */
//				poke16(0x62, led|0xf00); /* scl high, sda high */
//			}
//		}
//		poke16(0x62, led|0x800); /* scl low, tristate sda */
//		poke16(0x62, led|0xc00); /* scl low, sda low */
//		poke16(0x62, led|0xe00); /* scl high, sda low */
//		poke16(0x62, led|0x0); /* tristate (i2c stop) */
//		printf("rtc_oscillator_ok=%d\n", (rtcdat[1] & 0x80) ? 0 : 1);
//		// for (i = 0; i < 8; i++) printf("dat%d=0x%x\n", i, rtcdat[i]);
//		if ((rtcdat[1] & 0x80) == 0) {
//			tm.tm_sec = frombcd(rtcdat[0] & 0x7f);
//			tm.tm_min = frombcd(rtcdat[1] & 0x7f);
//			tm.tm_hour = frombcd(rtcdat[2] & 0x3f);
//			tm.tm_mday = frombcd(rtcdat[4] & 0x3f);
//			tm.tm_mon = frombcd(rtcdat[5] & 0x1f) - 1;
//			tm.tm_year = frombcd(rtcdat[6]) + 100;
//			setenv("TZ", "UTC", 1);
//			now = mktime(&tm);
//			assert(now != -1);
//			tv.tv_sec = now;
//			tv.tv_usec = 0;
//			settimeofday(&tv, NULL);
//		}
//	}
//
//skiprtc:

	/* Should be last! */
	if (opt_watchdog) {
		struct sched_param sched;
		int fd, r;
		char c;

		daemon(0, 0);
		sched.sched_priority = 98;
		sched_setscheduler(0, SCHED_FIFO, &sched);
		if (opt_autofeed) {
			unsigned int us = 60000000;
			if (wdog == 0) {
				us = 100000;
			} else if (wdog == 1) {
				us = 1000000;
			} else if (wdog == 2) {
				us = 5000000;
			}
			for (;;) {
				poke16(0x06, wdog);
				usleep(us);
			}
		}
		/* Single character writes to /dev/watchdog keep it fed */
		mknod("/dev/watchdog", S_IFIFO|0666, 0);
		for(;;) {
			fd = open("/dev/watchdog", O_RDONLY);
			assert (fd != -1);
			do {
				r = read(fd, &c, 1);
				if (c > '3' || c < '0') c = '3';
				poke16(0x06, c - '0');
			} while (r > 0);
			close(fd);
		}
	}


	return 0;
}


#define MPMU_BASE mpmu_base
#define MPMU_FCCR       (*((volatile unsigned int *)(MPMU_BASE + 0x08)))
#define MPMU_POSR       (*((volatile unsigned int *)(MPMU_BASE + 0x10)))
#define MPMU_PLL1REG1   (*((volatile unsigned int *)(MPMU_BASE + 0x50)))
#define MPMU_PLL2REG1   (*((volatile unsigned int *)(MPMU_BASE + 0x60)))

#define PMUA_BASE (0x2800+pmua_base)
#define APMU_CCR        (*((volatile unsigned int *)(PMUA_BASE + 0x4)))


static void calculateClocks(unsigned int *pclk, unsigned int *dclk)
{
   unsigned int vco_ref_div,
      vco_fb_div,
      vco_div_sel_diff,
      vco_div_sel_se,
      core_clk_div,
      ddr_clk_div,
      mpmu_fccr;

   unsigned int pll1_out, pll2_out, core_clk, ddr_clk;

   mpmu_fccr = MPMU_FCCR;
   MPMU_FCCR = mpmu_fccr | (1 << 14);       // Set PLL1CEN so that we can read PLL1 status

   switch ((MPMU_FCCR >> 29) & 3)
   {
      case 2:
      {
         //ser_puts("Core and DDR2 are using PLL2\n\r");

         vco_ref_div = (MPMU_POSR >> 23) & 0x1F;

         if (vco_ref_div != 1)
            fprintf(stderr, "Error:  APMU_CCR::PLL2REFD must be set to 1\n\r");
         else
         {
            vco_fb_div = (MPMU_POSR >> 14) & 0x1FF;
            vco_div_sel_diff = (MPMU_PLL2REG1 >> 19) & 0xF;
            vco_div_sel_se = (MPMU_PLL2REG1 >> 23) & 0xF;
            core_clk_div = APMU_CCR & 3;
            ddr_clk_div = (APMU_CCR >> 12) & 3;

            pll2_out = (26000 * vco_fb_div / 3) * 1000;

            core_clk = pll2_out / (core_clk_div + 1);

            if (vco_div_sel_se == 1)
               core_clk = core_clk * 2 / 3;
            else if (vco_div_sel_se == 2)
               core_clk >>= 1;

            ddr_clk = pll2_out / (2 * (ddr_clk_div+1));

            if (vco_div_sel_diff == 1)
               ddr_clk = ddr_clk * 2 / 3;    // divide by 1.5
            else if (vco_div_sel_diff == 2)
               ddr_clk >>= 1;                // divide by 2
            else if (vco_div_sel_diff == 3)
               ddr_clk = ddr_clk * 2 / 5;    // divide by 2.5
            else if (vco_div_sel_diff == 4)
               ddr_clk = ddr_clk * 2 / 6;    // divide by 3

            *pclk = core_clk / 1000000;
            *dclk = ddr_clk / 1000000;
         }
      }
      break;
      case 1:
      {
         //ser_puts("Core and DDR2 are using PLL1 @ 624MHz\n\r");

         pll1_out = 624000000;

         vco_div_sel_se = (MPMU_PLL1REG1 >> 23) & 0xF;

         core_clk_div = APMU_CCR & 3;

         ddr_clk_div = (APMU_CCR >> 12) & 3;

         core_clk = pll1_out / (core_clk_div + 1);
         if (vco_div_sel_se == 1)
            core_clk = core_clk * 2 / 3;
         else if (vco_div_sel_se == 2)
            core_clk >>= 1;

         ddr_clk = pll1_out / (2 * (ddr_clk_div+1));

         if (vco_div_sel_se == 1)
            ddr_clk = ddr_clk * 2 / 3;
         else if (vco_div_sel_se == 2)
               ddr_clk >>= 1;

         *pclk = core_clk / 1000000;
         *dclk = ddr_clk / 1000000;
      }

      break;

      case 0:
      {
         //ser_puts("Core and DDR2 are using PLL1 @ 312MHz\n\r");

         pll1_out = 312000000;

         vco_div_sel_se = (MPMU_PLL1REG1 >> 23) & 0xF;

         core_clk_div = APMU_CCR & 3;

         ddr_clk_div = (APMU_CCR >> 12) & 3;

         core_clk = pll1_out / (core_clk_div + 1);
         if (vco_div_sel_se == 1)
            core_clk = core_clk * 2 / 3;
         else if (vco_div_sel_se == 2)
            core_clk >>= 1;

         ddr_clk = pll1_out / (2 * (ddr_clk_div+1));

         if (vco_div_sel_se == 1)
            ddr_clk = ddr_clk * 2 / 3;
         else if (vco_div_sel_se == 2)
               ddr_clk >>= 1;

         *pclk = core_clk / 1000000;
         *dclk = ddr_clk / 1000000;
      }
      break;
   }

   MPMU_FCCR = mpmu_fccr;     // restore it

}
